/**
 * Copyright (c) 2009-2011 Zauber S.A. <http://www.zaubersoftware.com/>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.zauberlabs.commons.ws.connection;

import java.net.URL;
import java.util.Collection;
import java.util.LinkedList;
import java.util.Map;

import javax.xml.namespace.QName;
import javax.xml.ws.BindingProvider;
import javax.xml.ws.Service;
import javax.xml.ws.handler.soap.SOAPHandler;
import javax.xml.ws.handler.soap.SOAPMessageContext;

import com.zauberlabs.commons.ws.SoapHandlers;
import com.zauberlabs.commons.ws.security.UsernameTokenProfile;


/**
 * ConnectionBuilder
 * 
 * @author flbulgarelli
 * @param <A> the type of service built
 */
public class ConnectionBuilder<A> {

    private Class<? extends Service> serviceType;
    private Class<A> portType;
    private URL wsdlUrl;
    private String endpoint;
    private PortFactory portFactory = new Basic();
    private Collection<SOAPHandler<SOAPMessageContext>> headers = new LinkedList<SOAPHandler<SOAPMessageContext>>();
    private Collection<Configuration<? super A>> configurations = new LinkedList<Configuration<? super A>>();

    /**
     * Creates a {@link ConnectionBuilder}
     * @param portType
     */
    public ConnectionBuilder(final Class<A> portType) {
        this.portType = portType;
    }

    /** mandatory */
    public static <A> ConnectionBuilder<A> fromPortType(final Class<A> portType) {
        return new ConnectionBuilder<A>(portType);
    }

    /** mandatory */
    public ConnectionBuilder<A> withServiceType(final Class<? extends Service> serviceType) {
        this.serviceType = serviceType;
        return this;
    }

    /** mandatory when {@link #withClasspathWsdl(String)} is not used */
    public ConnectionBuilder<A> withWsdlUrl(final URL wsdlUrl) {
        this.wsdlUrl = wsdlUrl;
        return this;
    }

    /** mandatory when {@link #withWsdlUrl(URL)} is not used */
    public ConnectionBuilder<A> withClasspathWsdl(final String resourceName) {
        URL resource = Thread.currentThread().getContextClassLoader().getResource(resourceName);
        if (resource == null) {
            throw new IllegalArgumentException("Resource " + resourceName + " not found");
        }
        return withWsdlUrl(resource);
    }

    /** optional */
    public ConnectionBuilder<A> withEndpoint(final String endpoint) {
        this.endpoint = endpoint;
        return this;
    }

    /** optional */
    public ConnectionBuilder<A> withHeader(final SOAPHandler<SOAPMessageContext> header) {
        this.headers.add(header);
        return this;
    }

    /** optional */
    public ConnectionBuilder<A> withPortFactory(final PortFactory portFactory) {
        this.portFactory = portFactory;
        return this;
    }

    /** optional */
    public final ConnectionBuilder<A> withPortQName(final String qname) {
        return withPortQName(QName.valueOf(qname));
    }

    /** optional */
    public final ConnectionBuilder<A> withPortQName(final QName qname) {
        return withPortFactory(new WithQName(qname));
    }

    /** optional */
    public final ConnectionBuilder<A> withUsernameTokenAuth(final String username, final String password) {
        return withHeader(new UsernameTokenProfile(username, password));
    }

    /** optional */
    public ConnectionBuilder<A> withConfiguration(final Configuration<? super A> configuration) {
        configurations.add(configuration);
        return this;
    }

    public A build() {
        try {
            if (wsdlUrl == null) {
                throw new IllegalArgumentException("The wsdl Url or Classpath is missing");
            }
            Object service = createService(serviceType, wsdlUrl);
            A port = portFactory.createPort(service, portType);
            configureEndpoint(port);
            configureHeaders(port);
            configureHooks(port);
            return port;
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    protected void configureHooks(final A port) {
        for (Configuration<? super A> config : configurations) {
            config.configure(port);
        }
    }

    protected void configureHeaders(final A port) {
        for (SOAPHandler<SOAPMessageContext> header : headers) {
            SoapHandlers.add(provideBindings(port), header);
        }
    }

    protected void configureEndpoint(final A port) {
        if (endpoint != null) {
            Map<String, Object> requestContext = provideBindings(port).getRequestContext();
            requestContext.put(BindingProvider.ENDPOINT_ADDRESS_PROPERTY, endpoint);
        }
    }

    protected Object createService(final Class<? extends Service> serviceType, final URL wsdlUrl) throws Exception {
        return serviceType.getConstructor(URL.class).newInstance(wsdlUrl);
    }

    protected BindingProvider provideBindings(final A port) {
        return (BindingProvider) port;
    }

    /**
     * {@link PortFactory} that simply returns the result of sending
     * {@link Service#getPort(Class)}
     */
    static class Basic implements PortFactory {
        @Override
        public <A> A createPort(final Object service, final Class<A> portType) {
            return ((Service) service).getPort(portType);
        }
    }

    /**
     * {@link PortFactory} that returns the result of sendings
     * {@link Service#getPort(QName, Class)}
     */
    static class WithQName implements PortFactory {
        private QName portName;

        /**Creates a {@link WithQName} port factory*/
        public WithQName(final QName portName) {
            this.portName = portName;
        }

        @Override public <A> A createPort(final Object service, final Class<A> portType) {
            return ((Service) service).getPort(portName, portType);
        }
    }
}
